import java.awt.Color;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.TreeMap;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JRadioButton;
import javax.swing.text.Document;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;

public class Client extends javax.swing.JFrame {

    private final String[] smiles = {"B-)", ":@", ":-S", "<B", ":-P", ":-*", ":-D", ":'(", ":-(", ":-o"};
    private ObjectInputStream ois;
    private ObjectOutputStream oos;
    private String selectedUser, username;
    private final TreeMap<String, Object[]> onlineUsers = new TreeMap<>();
    private SourceDataLine sline;

    public Client(ObjectInputStream ois, ObjectOutputStream oos, String username) {
        initComponents();
        this.ois = ois;
        this.oos = oos;
        this.username = username;
        setTitle("Client: " + username);
        setVisible(true);
    }

    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        grp_users = new javax.swing.ButtonGroup();
        scr_conversation = new javax.swing.JScrollPane();
        txt_conversation = new javax.swing.JTextPane();
        jLabel1 = new javax.swing.JLabel();
        txt_message = new javax.swing.JTextField();
        btn_send = new javax.swing.JButton();
        jLabel2 = new javax.swing.JLabel();
        lyr_smile = new javax.swing.JLayeredPane();
        pnl_smile = new javax.swing.JPanel();
        pnl_setting = new javax.swing.JPanel();
        chk_blocksms = new javax.swing.JCheckBox();
        chk_blockvoice = new javax.swing.JCheckBox();
        scr_users = new javax.swing.JScrollPane();
        pnl_users = new javax.swing.JPanel();
        btn_voice = new javax.swing.JToggleButton();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("Server");
        setIconImage(new ImageIcon(getClass().getResource("xchat.png")).getImage());
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowOpened(java.awt.event.WindowEvent evt) {
                formWindowOpened(evt);
            }
        });

        txt_conversation.setEditable(false);
        txt_conversation.setName(""); // NOI18N
        scr_conversation.setViewportView(txt_conversation);

        jLabel1.setText("Message:");

        btn_send.setText("Send");
        btn_send.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                btn_sendActionPerformed(evt);
            }
        });

        jLabel2.setText("Smile:");

        lyr_smile.setLayout(new java.awt.GridLayout(1, 10, 0, 3));

        pnl_smile.setBorder(javax.swing.BorderFactory.createEmptyBorder(1, 1, 1, 1));
        pnl_smile.setLayout(new java.awt.GridLayout(1, 10, 0, 3));

        pnl_setting.setBorder(javax.swing.BorderFactory.createTitledBorder("Settings"));

        chk_blocksms.setText("Block SMS");
        chk_blocksms.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                chk_blocksmsActionPerformed(evt);
            }
        });

        chk_blockvoice.setText("Block Voice");
        chk_blockvoice.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                chk_blockvoiceActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout pnl_settingLayout = new javax.swing.GroupLayout(pnl_setting);
        pnl_setting.setLayout(pnl_settingLayout);
        pnl_settingLayout.setHorizontalGroup(
            pnl_settingLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(pnl_settingLayout.createSequentialGroup()
                .addComponent(chk_blocksms)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addComponent(chk_blockvoice)
                .addContainerGap())
        );
        pnl_settingLayout.setVerticalGroup(
            pnl_settingLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, pnl_settingLayout.createSequentialGroup()
                .addGap(0, 0, Short.MAX_VALUE)
                .addGroup(pnl_settingLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(chk_blocksms)
                    .addComponent(chk_blockvoice)))
        );

        scr_users.setBorder(javax.swing.BorderFactory.createEmptyBorder(2, 2, 2, 2));

        pnl_users.setBorder(javax.swing.BorderFactory.createTitledBorder("Users"));

        javax.swing.GroupLayout pnl_usersLayout = new javax.swing.GroupLayout(pnl_users);
        pnl_users.setLayout(pnl_usersLayout);
        pnl_usersLayout.setHorizontalGroup(
            pnl_usersLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 178, Short.MAX_VALUE)
        );
        pnl_usersLayout.setVerticalGroup(
            pnl_usersLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 365, Short.MAX_VALUE)
        );

        scr_users.setViewportView(pnl_users);

        btn_voice.setIcon(new javax.swing.ImageIcon(getClass().getResource("/Microphone-Hot-icon.png"))); // NOI18N
        btn_voice.setText("Voice");
        btn_voice.setDisabledSelectedIcon(null);
        btn_voice.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                btn_voiceActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(scr_conversation)
                    .addGroup(layout.createSequentialGroup()
                        .addGap(44, 44, 44)
                        .addComponent(lyr_smile))
                    .addGroup(layout.createSequentialGroup()
                        .addComponent(jLabel1)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(txt_message, javax.swing.GroupLayout.DEFAULT_SIZE, 225, Short.MAX_VALUE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(btn_send)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(btn_voice))
                    .addGroup(layout.createSequentialGroup()
                        .addComponent(jLabel2)
                        .addGap(3, 3, 3)
                        .addComponent(pnl_smile, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                    .addComponent(pnl_setting, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .addComponent(scr_users))
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(scr_conversation)
                    .addComponent(scr_users, javax.swing.GroupLayout.DEFAULT_SIZE, 362, Short.MAX_VALUE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(layout.createSequentialGroup()
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(txt_message, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(jLabel1)
                            .addComponent(btn_send)
                            .addComponent(btn_voice))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                            .addComponent(jLabel2, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .addComponent(pnl_smile, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)))
                    .addComponent(pnl_setting, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(lyr_smile, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addGap(0, 0, 0))
        );

        btn_voice.getAccessibleContext().setAccessibleName("");

        pack();
        setLocationRelativeTo(null);
    }// </editor-fold>//GEN-END:initComponents

    private void formWindowOpened(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowOpened
        new ClientThread(ois, oos).start();
        for (int i = 1; i <= 10; i++) {
            JLabel lbl = new JLabel(new ImageIcon(getClass().getResource(i + ".png")));
            lbl.addMouseListener(new SmileActionListener(i - 1));
            pnl_smile.add(lbl);
        }
        pnl_smile.revalidate();
        pnl_smile.repaint();
    }//GEN-LAST:event_formWindowOpened

    private void chk_blocksmsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_chk_blocksmsActionPerformed
        try {
            MailBox box = new MailBox();
            box.setUsername(username);
            box.setTitle("SmsBlock");
            Object[] info = new Object[]{selectedUser, chk_blocksms.isSelected()};
            box.setInfo(info);
            oos.writeObject(box);
        } catch (IOException ex) {
            JOptionPane.showMessageDialog(null, ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        }
    }//GEN-LAST:event_chk_blocksmsActionPerformed

    private void btn_sendActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_btn_sendActionPerformed
        try {
            MailBox box = new MailBox();
            box.setUsername(username);
            box.setTitle("Message");
            Object[] info = new Object[]{selectedUser, txt_message.getText()};
            box.setInfo(info);
            oos.writeObject(box);
            String user = "Me to " + selectedUser;
            InsertString(user, txt_message.getText() + "\n", Color.black);
            txt_message.setText("");
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(null, ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        }
    }//GEN-LAST:event_btn_sendActionPerformed

    private void chk_blockvoiceActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_chk_blockvoiceActionPerformed
        try {
            MailBox box = new MailBox();
            box.setUsername(username);
            box.setTitle("VoiceBlock");
            Object[] info = new Object[]{selectedUser, chk_blockvoice.isSelected()};
            box.setInfo(info);
            oos.writeObject(box);
        } catch (IOException ex) {
            JOptionPane.showMessageDialog(null, ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        }
    }//GEN-LAST:event_chk_blockvoiceActionPerformed

    private void btn_voiceActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_btn_voiceActionPerformed
        if (btn_voice.isSelected()) {
            AudioFormat af = getAudioFormat();
            try {
                sline = AudioSystem.getSourceDataLine(af);
                sline.open();
                sline.start();
            } catch (LineUnavailableException ex) {
            }
            VoiceThread voiceThread = new VoiceThread();
            voiceThread.start();
        } else {
            sline.stop();
            sline.close();
        }
    }//GEN-LAST:event_btn_voiceActionPerformed

    private class SmileActionListener extends MouseAdapter {

        private int index = 0;

        public SmileActionListener(int index) {
            this.index = index;
        }

        @Override
        public void mouseReleased(MouseEvent e) {
            txt_message.setText(txt_message.getText() + smiles[index]);
        }

    }

    private class VoiceThread extends Thread {

        @Override
        public void run() {
            try {
                MailBox box = new MailBox();
                box.setUsername(username);
                box.setTitle("Voice");
                AudioFormat af = getAudioFormat();
                TargetDataLine line = AudioSystem.getTargetDataLine(af);
                line.open();
                line.start();
                int frameSizeInBytes = af.getFrameSize();
                int bufferLengthInFrames = line.getBufferSize() / 8;
                int bufferLengthInBytes = bufferLengthInFrames * frameSizeInBytes;
                byte[] data = new byte[bufferLengthInBytes];
                int remaining = 0;
                while (btn_voice.isSelected()) {
                    remaining = line.read(data, 0, bufferLengthInBytes);
                    if (remaining == -1) {
                        break;
                    }
                    Object[] info = new Object[]{selectedUser, remaining, data};
                    box.setInfo(info);
                    oos.writeObject(box);
                    oos.reset();
                }
                line.stop();
                line.close();
            } catch (LineUnavailableException | IOException ex) {
                JOptionPane.showMessageDialog(null, ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }
        }

    }

    private class ClientThread extends Thread {

        private ObjectInputStream ois = null;
        private ObjectOutputStream oos = null;

        public ClientThread(ObjectInputStream ois, ObjectOutputStream oos) {
            this.oos = oos;
            this.ois = ois;
        }

        @Override
        public void run() {
            try {
                while (true) {
                    MailBox box = (MailBox) ois.readObject();
                    String title = box.getTitle();
                    switch (title) {
                        case "Message":
                            Message(ois, oos, box);
                            break;
                        case "Voice":
                            Voice(ois, oos, box);
                            break;
                        case "Users":
                            Users(ois, oos, box);
                            break;
                    }
                }
            } catch (IOException | ClassNotFoundException | LineUnavailableException ex) {
                JOptionPane.showMessageDialog(null, ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }
        }

    }

    private void Message(ObjectInputStream ois, ObjectOutputStream oos, MailBox box) {
        Object[] info = box.getInfo();
        String user = box.getUsername();
        String message = (String) info[1];
        InsertString(user, message + "\n", Color.black);
    }

    private void Voice(ObjectInputStream ois, ObjectOutputStream oos, MailBox box) throws LineUnavailableException, IOException {
        if (btn_voice.isSelected()) {
            Object[] info = box.getInfo();
            int remaining = (Integer) info[1];
            byte[] data = (byte[]) info[2];
            if (remaining != -1) {
                while (remaining > 0) {
                    remaining -= sline.write(data, 0, remaining);
                }
            }
        }
    }

    private void Users(ObjectInputStream ois, ObjectOutputStream oos, MailBox box) {
        pnl_users.removeAll();
        JRadioButton rdo_selected = null;
        chk_blocksms.setEnabled(false);
        chk_blockvoice.setEnabled(false);
        btn_send.setEnabled(false);
        btn_voice.setEnabled(false);
        Object[] onlinesUsers = box.getInfo();
        if (onlinesUsers.length <= 14) {
            pnl_users.setLayout(new GridLayout(14, 1));
        } else {
            pnl_users.setLayout(new GridLayout(0, 1));
        }
        for (Object tmp : onlinesUsers) {
            OnlineUser info = (OnlineUser) tmp;
            JRadioButton btn = new JRadioButton(info.getUsername());
            btn.addActionListener((ActionEvent e) -> {
                radioButtonClick(info, info.getUsername());
            });
            if (info.getUsername().equals(selectedUser)) {
                radioButtonClick(info, selectedUser);
                rdo_selected = btn;
            }
            grp_users.add(btn);
            pnl_users.add(btn);
        }
        pnl_users.repaint();
        pnl_users.revalidate();
        if (rdo_selected != null) {
            rdo_selected.setSelected(true);
        }
    }

    private void radioButtonClick(OnlineUser cls, String info) {
        chk_blocksms.setSelected(cls.isSmsBlock());
        chk_blockvoice.setSelected(cls.isVoiceBlock());
        selectedUser = info;
        chk_blocksms.setEnabled(true);
        chk_blockvoice.setEnabled(true);
        btn_send.setEnabled(true);
        btn_voice.setEnabled(true);
    }

    private synchronized void InsertString(String user, String msg, Color color) {
        try {
            Document doc = txt_conversation.getDocument();
            Style redStyle = txt_conversation.addStyle("redStyle", null);
            Style style = txt_conversation.addStyle("style", null);
            StyleConstants.setForeground(redStyle, Color.red);
            StyleConstants.setForeground(style, color);
            doc.insertString(doc.getLength(), "<" + user + ">: ", redStyle);
            ArrayList<String> msgMain = new ArrayList<>();
            msgMain.add(msg);
            for (int i = 0; i < smiles.length; i++) {
                ArrayList<String> msgSplit = new ArrayList<>();
                String smile = smiles[i];
                for (String tmp : msgMain) {
                    for (int j = 0; j < tmp.length();) {
                        int index = tmp.indexOf(smile, j);
                        if (index < 0) {
                            msgSplit.add(tmp.substring(j));
                            break;
                        }
                        msgSplit.add(tmp.substring(j, index));
                        msgSplit.add(tmp.substring(index, index + smile.length()));
                        j = index + smile.length();
                    }
                }
                msgMain = new ArrayList<>(msgSplit);
            }
            for (String tmp : msgMain) {
                int index = IndexOfSmile(tmp);
                if (index > 0) {
                    txt_conversation.setCaretPosition(doc.getLength());
                    txt_conversation.insertIcon(new ImageIcon(getClass().getResource(index + ".png")));
                } else {
                    doc.insertString(doc.getLength(), tmp, style);
                }
            }
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(null, ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        }
    }

    private AudioFormat getAudioFormat() {
        float sampleRate = 16000;
        int sampleSizeInBits = 8;
        int channels = 2;
        boolean signed = true;
        boolean bigEndian = true;
        AudioFormat format = new AudioFormat(sampleRate, sampleSizeInBits,
                channels, signed, bigEndian);
        return format;
    }

    private int IndexOfSmile(String str) {
        for (int i = 0; i < smiles.length; i++) {
            if (smiles[i].equals(str)) {
                return i + 1;
            }
        }
        return -1;
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton btn_send;
    private javax.swing.JToggleButton btn_voice;
    private javax.swing.JCheckBox chk_blocksms;
    private javax.swing.JCheckBox chk_blockvoice;
    private javax.swing.ButtonGroup grp_users;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JLayeredPane lyr_smile;
    private javax.swing.JPanel pnl_setting;
    private javax.swing.JPanel pnl_smile;
    private javax.swing.JPanel pnl_users;
    private javax.swing.JScrollPane scr_conversation;
    private javax.swing.JScrollPane scr_users;
    private javax.swing.JTextPane txt_conversation;
    private javax.swing.JTextField txt_message;
    // End of variables declaration//GEN-END:variables
}
